;;; ggltr-mode.el --- Wrapper of google-translate minor mode -*- lexical-binding: t -*-
;;;
;;; Version: 2024/12/07
;;;
;;; Install
;;;  copy ggltr-mode.el to /usr/share/emacs/site-lisp/.
;;;
;;; Depended packages
;;;  google-translate
;;;  popup
;;;
;;; Setting
;;;  add to .emacs
;;;  ----
;;;  (require 'ggltr-mode)
;;;  ----
;;;
;;; Start ggltr-mode
;;;  M-x ggltr-mode
;;;
;;; Other commands
;;;    M-x ggltr-set-source-language
;;;    M-x ggltr-set-target-language
;;;    M-x ggltr-kill-ring-save-trstring
;;;    M-x ggltr-toggle-auto-kill-ring-save
;;;    M-x ggltr-enable-auto-kill-ring-save
;;;    M-x ggltr-disable-auto-kill-ring-save
;;;    M-x ggltr-toggle-auto-target-select
;;;

;;; Code:

(require 'google-translate)
(require 'popup)

(defcustom ggltr-default-src-lang    "auto"    "ggltr-default-src-lang"    :group 'ggltr :type 'string)
(defcustom ggltr-default-tgt-lang    "ja"      "ggltr-default-tgt-lang"    :group 'ggltr :type 'string)
(defcustom ggltr-mode-line-str       " tr"     "ggltr-mode-line-str"       :group 'ggltr :type 'string)
(defcustom ggltr-input-idle-delay    0.8       "ggltr-input-idle-delay"    :group 'ggltr :type 'number)
(defcustom ggltr-auto-kill-ring-save nil       "ggltr-auto-kill-ring-save" :group 'ggltr :type 'boolean)
(defcustom ggltr-auto-target-select  nil       "ggltr-auto-target-select"  :group 'ggltr :type 'boolean)
(defcustom ggltr-default-tgt2-lang   "en"      "ggltr-default-tgt2-lang"   :group 'ggltr :type 'string)
(defcustom ggltr-tgt2-sel-src-chars  "[ぁ-んァ-ヴ一-龠ー々]"
  "ggltr-tgt2-sel-src-chars"   :group 'ggltr :type 'string)

(defvar ggltr-inuse nil "ggltr-inuse")
(make-variable-buffer-local 'ggltr-inuse)
(put 'ggltr-inuse 'permanent-local t)

(defvar ggltr-timer-object     nil "ggltr-timer-object")
(defvar ggltr-timer-refcounter 0   "ggltr-timer-refcounter")

(defvar ggltr-local-src-lang nil "ggltr-local-src-lang")
(make-variable-buffer-local 'ggltr-local-src-lang)
(put 'ggltr-local-src-lang 'permanent-local t)

(defvar ggltr-local-tgt-lang nil "ggltr-local-tgt-lang")
(make-variable-buffer-local 'ggltr-local-tgt-lang)
(put 'ggltr-local-tgt-lang 'permanent-local t)

(defvar ggltr-local-tgt2-lang nil "ggltr-local-tgt2-lang")
(make-variable-buffer-local 'ggltr-local-tgt2-lang)
(put 'ggltr-local-tgt2-lang 'permanent-local t)

(defvar ggltr-trstring nil "ggltr-trstring")


(define-minor-mode ggltr-mode
  "Toggle ggltr-mode.
ggltr-mode is a buffer-local minor mode."
  :lighter ggltr-mode-line-str
  :append-arg-docstring t

  (when ggltr-mode
    (setq ggltr-local-src-lang ggltr-default-src-lang)
    (setq ggltr-local-tgt-lang ggltr-default-tgt-lang)
    (setq ggltr-local-tgt2-lang ggltr-default-tgt2-lang)
    (ggltr-schedule)
    (add-hook 'kill-buffer-hook 'ggltr-cancel nil t)
    )
  (unless ggltr-mode
    (ggltr-cancel)
    (remove-hook 'kill-buffer-hook t)
    )
  )

(defun ggltr-set-source-language (&optional src-lang)
  "Set the translation source language."
  (interactive)
  (unless src-lang
    (let ((langlist '("AUTO")))
      (dolist (el google-translate-supported-languages-alist)
        (push (car el) langlist))
      (setq src-lang (completing-read (format "Source[%s]: " ggltr-local-src-lang) langlist))))
  (let ((short-lang (cdr (assoc src-lang google-translate-supported-languages-alist))))
    (if short-lang
        (setq ggltr-local-src-lang short-lang)
      (setq ggltr-local-src-lang "auto"))
    (message (format "set source-lang \"%s\"" ggltr-local-src-lang))
    ))

(defun ggltr-set-target-language (&optional tgt-lang)
  "Set the translation target language."
  (interactive)
  (unless tgt-lang
    (let ((langlist '()))
      (dolist (el google-translate-supported-languages-alist)
        (push (car el) langlist))
      (setq tgt-lang (completing-read (format "Target[%s]: " ggltr-local-tgt-lang) langlist))))
  (let ((short-lang (cdr (assoc tgt-lang google-translate-supported-languages-alist))))
    (when short-lang
      (setq ggltr-local-tgt-lang short-lang)
      (message (format "set target-lang \"%s\"" ggltr-local-tgt-lang))
      )))

(defun ggltr-update-current ()
  (when (and ggltr-inuse
             (region-active-p))
    (let* ((txtbuf (replace-regexp-in-string
                    "\n" " "
                    (funcall region-extract-function nil)))
           (seltgt2 (and ggltr-auto-target-select
                         (< (length (replace-regexp-in-string
                                     ggltr-tgt2-sel-src-chars ""
                                     txtbuf))
                            (length txtbuf)))))
      (ggltr-translate ggltr-local-src-lang
                       (if seltgt2
                           ggltr-local-tgt2-lang
                         ggltr-local-tgt-lang)
                       txtbuf
                       'popup)
      )))

(defun ggltr-schedule ()
  (setq ggltr-timer-refcounter (+ ggltr-timer-refcounter 1))
  (setq ggltr-inuse t)
  (when (eq ggltr-timer-refcounter 1)
    (setq ggltr-timer-object
          (run-with-idle-timer ggltr-input-idle-delay t #'ggltr-update-current))
    ))

(defun ggltr-cancel ()
  (setq ggltr-timer-refcounter (- ggltr-timer-refcounter 1))
  (setq ggltr-inuse nil)
  (when (and ggltr-timer-object
             (eq ggltr-timer-refcounter 0))
    (cancel-timer ggltr-timer-object)
    (setq ggltr-timer-object nil)
    ))

(defun ggltr-translate (source-language target-language text &optional output-destination)
  "Translate TEXT from SOURCE-LANGUAGE to TARGET-LANGUAGE."

  (let* ((json (google-translate-request source-language
                                         target-language
                                         text)))
    (if (null json)
        (message "Nothing to translate.")
      (let* ((detailed-translation
              (google-translate-json-detailed-translation json))
             (detailed-definition
              (google-translate-json-detailed-definition json))
             (gtos
              (make-gtos
               :source-language source-language
               :target-language target-language
               :auto-detected-language (aref json 2)
               :text text
               :text-phonetic (google-translate-json-text-phonetic json)
               :translation (google-translate-json-translation json)
               :translation-phonetic (google-translate-json-translation-phonetic json)
               :detailed-translation detailed-translation
               :detailed-definition detailed-definition
               :suggestion (when (null detailed-translation)
                             (google-translate-json-suggestion json))))
             (output-destination (if (null output-destination)
                                     google-translate-output-destination
                                   output-destination)))
        (cond
         ((equal output-destination 'popup)
          (ggltr-popup-output-translation gtos))
         )
        ))))

(defun ggltr-popup-output-translation (gtos)
  "Output translation to the popup tooltip using `popup'
package."
  (require 'popup)
  (popup-tip
   (with-temp-buffer
     (ggltr-insert-translation gtos)
     (setq ggltr-trstring (buffer-substring (point-min) (point-max)))
     (if ggltr-auto-kill-ring-save
         (kill-new ggltr-trstring))
     (google-translate--trim-string ggltr-trstring)
     )))

(defun ggltr-insert-translation (gtos)
  "Insert translation to the current buffer."
  (insert
   (google-translate--translated-text gtos "%s")
   ))

(defun ggltr-kill-ring-save-trstring ()
  "Save translation string to kill-ring."
  (interactive)
  (kill-new ggltr-trstring)
  )

(defun ggltr-toggle-auto-kill-ring-save ()
  "Toggle auto-kill-ring-save status."
  (interactive)
  (setq ggltr-auto-kill-ring-save
        (not ggltr-auto-kill-ring-save))
  (message "ggltr-auto-kill-ring-save %s"
           (if ggltr-auto-kill-ring-save
               "enabled."
             "disabled."))
  )

(defun ggltr-enable-auto-kill-ring-save ()
  "Enable auto-kill-ring-save."
  (interactive)
  (setq ggltr-auto-kill-ring-save t)
  (message "ggltr-auto-kill-ring-save %s" "enabled.")
  )

(defun ggltr-disable-auto-kill-ring-save ()
  "Disable auto-kill-ring-save."
  (interactive)
  (setq ggltr-auto-kill-ring-save nil)
  (message "ggltr-auto-kill-ring-save %s" "disabled.")
  )

(defun ggltr-toggle-auto-target-select ()
  "Toggle auto target select."
  (interactive)
  (setq ggltr-auto-target-select
        (not ggltr-auto-target-select))
  (message "ggltr-auto-target-select %s"
           (if ggltr-auto-target-select
               "enabled."
             "disabled."))
  )

(provide 'ggltr-mode)
